package sawat.tiling;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.FlowLayout;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;

import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JToggleButton;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileNameExtensionFilter;

public class Tessellation2 extends JComponent {
	private final class ColorIcon implements Icon {
		private final int i;

		private ColorIcon(int i) {
			this.i = i;
		}

		public int getIconHeight() {
			return 17;
		}

		public int getIconWidth() {
			return 50;
		}

		public void paintIcon(Component c, Graphics g, int x, int y) {
			g.setColor(i == 0 ? getMainColor() : i == 1 ? getSubColor()
					: getLineColor());
			g.fillRect(x, y, getIconWidth(), getIconHeight());
		}
	}

	// {ƂȂlp`
	private Polygon shape;
	// {ƂȂlp`hF
	private Color mainColor;
	// 180x]lp`hF
	private Color subColor;

	// E`F
	private Color lineColor;
	// ẼXg[N
	private BasicStroke stroke;
	// E`L
	private boolean drawLine = true;
	// Of[VL
	private boolean gradient = true;
	// A`GCAX
	private boolean antialias = true;

	// 菑lp`
	private ArrayList<Point> freePoints = new ArrayList<Point>(4);

	private final JComponent control;
	private JToggleButton editButton;

	/**
	 * C
	 * 
	 * @param args
	 */
	public static void main(String[] args) {

		if (args.length == 0) {
			java.awt.EventQueue.invokeLater(new Runnable() {
				public void run() {
					showWindow();
				}
			});
			return;
		}

		Tessellation2 t = new Tessellation2();
		t.setSize(600, 600);
		for (int i = 0; i < args.length; i++) {
			t.reflesh();
			t.saveImage(new File(args[i]));
		}
	}

	/**
	 * 
	 */
	public static void showWindow() {
		JFrame f = new JFrame("ʏ[Ulp`");
		f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		final Tessellation2 tessellation = new Tessellation2();
		f.getContentPane().add(tessellation, BorderLayout.CENTER);
		f.getContentPane().add(tessellation.getControl(), BorderLayout.EAST);
		f.pack();
		f.setLocationRelativeTo(null);
		f.setVisible(true);
	}

	/** RXgN^ */
	public Tessellation2() {
		this.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseReleased(MouseEvent e) {
				if (isEditMode()) {
					addPoint(e.getPoint());
				} else {
					// }EXNbNŐ}ƐFω
					//reflesh();
					//Tessellation.this.getControl().repaint();
					showIJ(e.getPoint());
				}
				Tessellation2.this.repaint();
			}
		});
		stroke = new BasicStroke(1.0f);
		reflesh();

		this.control = createControlPanel();
	}

	@Override
	public Dimension getPreferredSize() {
		return new Dimension(400, 300);
	}

	protected void addPoint(Point newPoint) {
		freePoints.add(newPoint);
		System.out.println(newPoint);
		if (freePoints.size() == 4) {
			final Polygon pol = new Polygon();
			for (Point p : freePoints) {
				pol.addPoint(p.x, p.y);
			}
			Rectangle b = pol.getBounds();

			final Polygon pol2 = new Polygon();
			for (Point p : freePoints) {
				pol2.addPoint(p.x - (int) b.getCenterX(), p.y
						- (int) b.getCenterY());
			}

			EventQueue.invokeLater(new Runnable() {
				public void run() {
					shape = pol2;
					freePoints.clear();
					editButton.setSelected(false);
				}
			});
		}
	}

	/**
	 * @return
	 */
	private JComponent createControlPanel() {

		JPanel control = new JPanel(new FlowLayout(FlowLayout.LEADING));
		control.setBorder(BorderFactory.createEtchedBorder(1));
		final Dimension dem = Tessellation2.this.getPreferredSize();
		dem.width = 100;
		{
			final JCheckBox check = new JCheckBox("");
			check.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					drawLine = check.isSelected();
					Tessellation2.this.repaint();
				}
			});
			check.setSelected(drawLine);
			control.add(check);
			dem.width = Math.max(dem.width, check.getWidth());
		}
		{
			final JSlider slider = new JSlider(0,100,15);
			slider.setPreferredSize(new Dimension(50, 30));
			slider.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					int v = slider.getValue();
					stroke = new BasicStroke((float)v/10);
					Tessellation2.this.repaint();
				}
			});
			control.add(slider);
		}
		control.setPreferredSize(dem);
		{
			final JCheckBox check = new JCheckBox("Of[V");
			check.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					gradient = check.isSelected();
					Tessellation2.this.repaint();
				}
			});
			check.setSelected(gradient);
			control.add(check);
			dem.width = Math.max(dem.width, check.getPreferredSize().width);
		}
		{
			final JCheckBox check = new JCheckBox("A`GCAX");
			check.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					antialias = check.isSelected();
					Tessellation2.this.repaint();
				}
			});
			check.setSelected(antialias);
			control.add(check);
			dem.width = Math.max(dem.width, check.getPreferredSize().width);
		}
		{
			editButton = new JToggleButton("菑");
			editButton.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent e) {
					Tessellation2.this.repaint();
				}
			});
			editButton.setSelected(false);
			control.add(editButton);
			dem.width = Math.max(dem.width, editButton.getWidth());
		}
		{
			final JButton button = new JButton("_");
			button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					shape = createShape();
					Tessellation2.this.repaint();
				}
			});
			control.add(button);
			dem.width = Math.max(dem.width, button.getPreferredSize().width);
		}
		{
			final JButton button = new JButton(new ColorIcon(0));
			button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					final Color newColor = JColorChooser.showDialog(button,
							"F1I", mainColor);
					if (newColor != null) {
						mainColor = newColor;
						Tessellation2.this.repaint();
					}
				}
			});
			control.add(button);
			dem.width = Math.max(dem.width, button.getPreferredSize().width);
		}
		{
			final JButton button = new JButton(new ColorIcon(1));
			button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					final Color newColor = JColorChooser.showDialog(button,
							"F2I", mainColor);
					if (newColor != null) {
						subColor = newColor;
						Tessellation2.this.repaint();
					}
				}
			});
			control.add(button);
			dem.width = Math.max(dem.width, button.getPreferredSize().width);
		}
		{
			final JButton button = new JButton(new ColorIcon(2));
			button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					final Color newColor = JColorChooser.showDialog(button,
							"F3I", mainColor);
					if (newColor != null) {
						lineColor = newColor;
						Tessellation2.this.repaint();
					}
				}
			});
			control.add(button);
			dem.width = Math.max(dem.width, button.getPreferredSize().width);
		}
		{
			final JButton button = new JButton("ۑ");
			button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					FileNameExtensionFilter filter = new FileNameExtensionFilter(
							"PNG / JPG / GIF / BMP", "jpg", "gif", "png", "bmp");

					JFileChooser fc = new JFileChooser("image.png");
					fc.setFileFilter(filter);

					fc.setDialogType(JFileChooser.SAVE_DIALOG);
					int i = fc.showDialog(button, null);
					if (i == JFileChooser.APPROVE_OPTION) {
						boolean b = saveImage(fc.getSelectedFile());
						if (!b) {
							JOptionPane.showMessageDialog(button,
									"ۑɎs܂B\n" + fc.getSelectedFile(),
									"ۑs", JOptionPane.ERROR_MESSAGE);
						}
					}
				}
			});
			control.add(button);
			dem.width = Math.max(dem.width, button.getPreferredSize().width);
		}
		dem.width += 20;
		control.setPreferredSize(dem);

		return control;
	}

	/**
	 * Vlp`B
	 * 
	 * @return lp`PolygonIuWFNg
	 */
	private Polygon createShape() {

		int size = 150;

		Random rand = new Random();
		int[] xs = new int[4];
		int[] ys = new int[4];
		double d = 0.5 * Math.PI * rand.nextDouble();
		double d1 = Math.PI * rand.nextDouble();
		double d2 = Math.PI - d1;

		for (int i = 0; i < 4; i++) {
			final double r = rand.nextDouble();
			xs[i] = (int) (Math.cos(d) * (size * r));
			ys[i] = (int) (Math.sin(d) * (size * r));

			d += (i % 2 == 0) ? d1 : d2;
			System.out.printf("->(%d,%d)", xs[i], ys[i]);
		}
		System.out.println();

		return new Polygon(xs, ys, 4);
	}

	/** ʏ[U̕`揈 */
	@Override
	protected void paintComponent(Graphics _g) {
		Graphics2D g = (Graphics2D) _g;
		if (antialias) {
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
					RenderingHints.VALUE_ANTIALIAS_ON);
		}

		int w = getWidth();
		int h = getHeight();
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, w, h);

		if (isEditMode()) {
			g.setColor(Color.BLACK);
			Point pp = null;
			for (Point p : freePoints) {
				g.drawRect(p.x - 1, p.y - 1, 3, 3);
				if (pp != null) {
					g.drawLine(pp.x, pp.y, p.x, p.y);
				}
				pp = p;
			}
		} else {
			g.translate(w / 2, h / 2);
			fillRepeate(g, shape);
		}
	}

	/**
	 * }`~l߂B
	 * 
	 * @param g
	 * @param s
	 */
	private void fillRepeate(Graphics2D g, Polygon s) {

		int xd0 = s.xpoints[0] - s.xpoints[2];
		int yd0 = s.ypoints[0] - s.ypoints[2];
		int xd1 = s.xpoints[1] - s.xpoints[3];
		int yd1 = s.ypoints[1] - s.ypoints[3];

		// 180x]
		Shape rev = AffineTransform.getQuadrantRotateInstance(2)
				.createTransformedShape(s);
		// Ήӂׂ荇ʒu֕sړ
		rev = AffineTransform.getTranslateInstance(s.xpoints[0] + s.xpoints[1],
				s.ypoints[0] + s.ypoints[1]).createTransformedShape(rev);

		g.setStroke(stroke);

		final int w = getWidth();
		final int h = getHeight();
		int ni = 1 + repeatCount(w, h, xd0, yd0);
		int nj = 1 + repeatCount(w, h, xd1, yd1);

		for (int i = -ni; i <= ni; i++) {
			for (int j = -nj; j <= nj; j++) {
				Shape ss = AffineTransform.getTranslateInstance(
						xd0 * i + xd1 * j, yd0 * i + yd1 * j)
						.createTransformedShape(s);
				Shape revSS = AffineTransform.getTranslateInstance(
						xd0 * i + xd1 * j, yd0 * i + yd1 * j)
						.createTransformedShape(rev);
				g.setPaint(getPaint(ss, true));
				g.fill(ss);
				g.setPaint(getPaint(revSS, false));
				g.fill(revSS);

				if (drawLine) {
					g.setPaint(lineColor);
					g.draw(ss);
					g.drawString(String.format("%d,%d",i, j), xd0 * i + xd1 * j, yd0 * i + yd1 * j);
				}
			}
		}
	}

	/**
	 * @param s
	 * @param main
	 * @return
	 */
	private Paint getPaint(Shape s, boolean main) {
		if (!gradient) return (main ? mainColor : subColor);

		Rectangle2D b = s.getBounds2D();

		return new GradientPaint((float) b.getMinX(), (float) b.getMinY(),
				mainColor, (float) b.getMaxX(), (float) b.getMaxY(), subColor);
	}

	/**
	 * JԂ񐔂߂iKɁj
	 * 
	 * @param w
	 * @param w
	 * @param h
	 * @param xd
	 * @param yd
	 * @return
	 */
	private int repeatCount(int w, int h, int xd, int yd) {

		return xd == 0 ? h / Math.abs(yd) : yd == 0 ? w / Math.abs(xd) : Math
				.max(w / Math.abs(xd), h / Math.abs(yd));
	}

	/**
	 * lp`ƐF_ĐB
	 */
	private void reflesh() {
		shape = createShape();

		mainColor = new Color(Color.HSBtoRGB((float) Math.random(), 1.0f, 1.0f));
		subColor = new Color(Color.HSBtoRGB((float) Math.random(), 0.7f, 1.0f));
		lineColor = new Color(Color.HSBtoRGB((float) Math.random(),
				(float) Math.random(), (float) Math.random()));
	}

	protected void showIJ(Point p) {
		Polygon s = shape;
		int w = getWidth(), h = getHeight();
		double x = p.x - w/2 , y = p.y - h/2;
		int xd0 = s.xpoints[0] - s.xpoints[2];
		int yd0 = s.ypoints[0] - s.ypoints[2];
		int xd1 = s.xpoints[1] - s.xpoints[3];
		int yd1 = s.ypoints[1] - s.ypoints[3];
		
		double j = (xd0 * y - yd0 * x) / (xd0 * yd1 - yd0 * xd1);
		double i = (x - xd1 * j) / xd0;
//		System.out.printf("%3.2f, %3.2f%n", i, j);
		System.out.printf("%d, %d%n", Math.round(i), Math.round(j));
	}
	
	public boolean isDrawLine() {
		return this.drawLine;
	}

	public void setDrawLine(boolean drawLine) {
		this.drawLine = drawLine;
	}

	public Color getLineColor() {
		return this.lineColor;
	}

	public void setLineColor(Color lineColor) {
		this.lineColor = lineColor;
	}

	public void setMainColor(Color mainColor) {
		this.mainColor = mainColor;
	}

	public Color getSubColor() {
		return this.subColor;
	}

	public void setSubColor(Color subColor) {
		this.subColor = subColor;
	}
	public Color getMainColor() {
		return this.mainColor;
	}

	public JComponent getControl() {
		return control;
	}

	public boolean saveImage(File f) {
		System.out.println(f);
		BufferedImage image = new BufferedImage(getWidth(), getHeight(),
				BufferedImage.TYPE_INT_BGR);
		Graphics g = image.getGraphics();
		paintComponent(g);

		try {
			final String type = f.getName().replaceFirst("^[^.]*.(.*)$", "$1");

			return ImageIO.getImageWritersBySuffix(type).hasNext()
					&& ImageIO.write(image, type, f);
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}
	}

	public boolean isEditMode() {
		return editButton.isSelected();
	}

}
